#ifndef ATTRIBUTEITEM_H
#define ATTRIBUTEITEM_H

#include <QStringList>
#include <QVariant>

#include "treeitem.h"

/*! \brief A key and a value 
    The data itself is stored in Attribute Definitions (AttributeDef). 
	A list of these tables AttributeTable is maintained for every MapEditor.
*/
class AttributeItem:public TreeItem {

enum Type {
	Undefined,	//!< Undefined type
	IntList,	//!< Free integer
	FreeInt,	//!< Free integer
	StringList, //!< List of strings, one can be attribute value
	FreeString,	//!< Any string can be attribute value, not unique
	UniqueString//!< UniqueString, e.g. for IDs
};

public:
	AttributeItem(const QList<QVariant> &data, TreeItem *parent = 0);
	virtual ~AttributeItem();
	void setKey (const QString &k, const Type &t);
	QString getKey ();
	void setValue (const QString &v);
	QVariant getValue ();
	void setType (const Type &t);
	AttributeItem::Type getAttributeType ();
	QString getTypeString ();
	QString getDataXML();
protected:
	QString freeString;		//!< String value for type FreeString
	bool internal;			//!< Internal attributes cannot be edited by user
};

#endif

