#ifndef BRANCHITEM_H
#define BRANCHITEM_H

#include "mapitem.h"


class QString;
class QGraphicsScene;
class BranchObj;
class XLinkItem;

class BranchItem:public MapItem
{
public:
    BranchItem(const QList<QVariant> &data, TreeItem *parent = 0);
    virtual ~BranchItem();
	virtual void copy (BranchItem *item);
	virtual BranchItem* parentBranch();

	virtual void insertBranch (int pos,BranchItem *branch);

	virtual QString saveToDir (const QString &tmpdir,const QString &prefix, const QPointF& offset);

	virtual void updateVisibility();

	virtual void setHeadingColor (QColor color); //!Overloaded from TreeItem to update QGraphicsView

protected:	
	bool scrolled;			// true if all children are scrolled and thus invisible
	bool tmpUnscrolled;		// can only be true (temporary) for a scrolled subtree
public:
	virtual void unScroll();				
	virtual bool toggleScroll();			// scroll or unscroll
	virtual bool isScrolled();				// returns scroll state
	virtual bool hasScrolledParent(BranchItem*);	// true, if any of the parents is scrolled
	virtual bool tmpUnscroll();				// unscroll scrolled parents temporary e.g. during "find" process
	virtual bool resetTmpUnscroll();		// scroll all tmp scrolled parents again e.g. when unselecting
	virtual void sortChildren();		//! Sort children 

protected:
	bool includeImagesVer;			//! include floatimages in bbox vertically
	bool includeImagesHor;			//! include floatimages in bbox horizontally
public:
	void setIncludeImagesVer(bool);
	bool getIncludeImagesVer();
	void setIncludeImagesHor(bool);
	bool getIncludeImagesHor();
	QString getIncludeImageAttr();

protected:
	int lastSelectedBranchNum;
	int lastSelectedBranchNumAlt;
public:
	virtual void setLastSelectedBranch();			//! Set myself as last selected in parent
	virtual void setLastSelectedBranch(int i);		//! Set last selected branch directly
	virtual BranchItem* getLastSelectedBranch();	//! Returns last selected branch usually
	virtual BranchItem* getLastSelectedBranchAlt(); //! Used to return last selected branch left of a mapcenter

public:
	TreeItem* findMapItem (QPointF p,TreeItem* excludeTI);	//! search map for branches or images. Ignore excludeTI, where search is started 
//	virtual TreeItem* findID (QString sid);	//! search map for object with ID string

	virtual void updateStyles ();			//! update related fonts, parObjects, links, ...
	virtual BranchObj* getBranchObj();	
	virtual BranchObj* createMapObj(QGraphicsScene *scene);		//! Create classic object in GraphicsView
};

#endif
