#ifndef BRANCHOBJ_H
#define BRANCHOBJ_H

#include "floatimageobj.h"
#include "linkablemapobj.h"
#include "ornamentedobj.h"
#include "xlinkobj.h"

bool isAbove(BranchObj*,BranchObj*);

/*! \brief A branch visible in the map */

/////////////////////////////////////////////////////////////////////////////
class BranchObj:public OrnamentedObj {
public:
	/*! New branches will get use same color for heading as parent */
	enum BranchModification {NewBranch, MovedBranch};

    BranchObj (QGraphicsScene*,TreeItem *ti=NULL);
    ~BranchObj ();
	bool operator< ( const BranchObj & );
	bool operator== ( const BranchObj & );
    virtual void init ();
    virtual void copy (BranchObj*);
    void clear();

    virtual void setParObjTmp (LinkableMapObj*,QPointF,int);// Only for moving Obj around
	virtual void unsetParObjTmp();			// reuse original ParObj

	virtual void setVisibility(bool,int);	// set visibility
    virtual void setVisibility(bool);	    // set vis. for w
	virtual void setLinkColor();			// set the color of link

	virtual void updateContentSize();
	virtual void positionContents();
    virtual void move (double x,double y);
    virtual void move (QPointF);
    virtual void moveBy (double x,double y);
    virtual void moveBy (QPointF);
    virtual void positionBBox();
    virtual void calcBBoxSize();
	virtual void setDockPos();
    
    virtual void updateData();	//! Update represantatio of heading, flags, etc.

protected:	
	virtual void savePosInAngle();					// write pos in angle for resorting			
public:	
	virtual void setDefAttr (BranchModification);	// set default attributes (font, size, ...)

    virtual void alignRelativeTo(const QPointF, bool alignSelf=false );
	virtual void reposition();
	virtual void unsetAllRepositionRequests();

	virtual QRectF getTotalBBox();			// return BBox including children			
	virtual QRectF getBBoxSizeWithChildren();	// return size of BBox including children  
	virtual ConvexPolygon getBoundingPolygon();
	virtual void calcBBoxSizeWithChildren();	// calc size of  BBox including children recursivly

	virtual QString getSelectString();
	virtual void setAnimation(const AnimPoint &ap);
	virtual bool animate();

protected:
	QList<XLinkObj*> xlink;			// xlinks to other branches

	AnimPoint anim;

public:	
	float angle;					// used in mainbranch to reorder mainbranches
};


#endif

