#include "ornamentedobj.h"
#include "linkablemapobj.h"
#include "vymmodel.h"

/////////////////////////////////////////////////////////////////
// OrnamentedObj
/////////////////////////////////////////////////////////////////


OrnamentedObj::OrnamentedObj(QGraphicsScene* s,TreeItem *ti) :LinkableMapObj(s,ti)
{
//    cout << "Const OrnamentedObj (s)\n";
    init ();
}

OrnamentedObj::OrnamentedObj (OrnamentedObj* lmo) : LinkableMapObj (lmo->scene)
{
    copy (lmo);
}

OrnamentedObj::~OrnamentedObj()
{
    delete heading;
	delete systemFlags;
	delete standardFlags;
	delete frame;
}


void OrnamentedObj::init ()
{
	heading = new HeadingObj(scene);
	heading->move (absPos.x(), absPos.y());

	systemFlags=new FlagRowObj(scene);
	standardFlags=new FlagRowObj(scene);

	frame = new FrameObj (scene);
}

void OrnamentedObj::copy (OrnamentedObj* other)
{
    LinkableMapObj::copy(other);
	heading->copy(other->heading);
    setColor   (other->heading->getColor());	

	systemFlags->copy (other->systemFlags);
	standardFlags->copy (other->standardFlags);

	ornamentsBBox=other->ornamentsBBox;
}

void OrnamentedObj::setLinkColor()
{
	VymModel *model=treeItem->getModel();
	if (!model) return;
	if (model->getMapLinkColorHint()==HeadingColor)
		LinkableMapObj::setLinkColor (heading->getColor());
	else	
		LinkableMapObj::setLinkColor (model->getMapDefLinkColor());
}

void OrnamentedObj::setColor (QColor col)
{
    heading->setColor(col);
	setLinkColor();
}

QColor OrnamentedObj::getColor ()
{
    return heading->getColor();
}

FrameObj* OrnamentedObj::getFrame()
{
	return frame;
}

FrameObj::FrameType OrnamentedObj::getFrameType()
{
	return frame->getFrameType();
}

QString OrnamentedObj::getFrameTypeName()
{
	return frame->getFrameTypeName();
}

void OrnamentedObj::setFrameType(const FrameObj::FrameType &t)
{
	frame->setFrameType(t);
	if (t == FrameObj::NoFrame)
		linkpos=LinkableMapObj::Bottom;
	else	
		linkpos=LinkableMapObj::Middle;

	calcBBoxSize();
	positionBBox();
	requestReposition();
}

void OrnamentedObj::setFrameType(const QString &t)
{
	frame->setFrameType(t);
	if (frame->getFrameType() == FrameObj::NoFrame)
		linkpos=LinkableMapObj::Bottom;
	else	
		linkpos=LinkableMapObj::Middle;

	calcBBoxSize();
	positionBBox();
	requestReposition();
}

void OrnamentedObj::setFramePadding (const int &i)
{
	frame->setPadding (i);
	calcBBoxSize();
	positionBBox();
	requestReposition();
}

int OrnamentedObj::getFramePadding ()
{
	return frame->getPadding();
}

void OrnamentedObj::setFrameBorderWidth (const int &i)
{
	frame->setBorderWidth(i);
	calcBBoxSize();
	positionBBox();
	requestReposition();
}

int OrnamentedObj::getFrameBorderWidth()
{
	return frame->getBorderWidth();
}

void OrnamentedObj::setFramePenColor(QColor col)
{
	frame->setPenColor (col);
}

QColor OrnamentedObj::getFramePenColor()
{
	return frame->getPenColor ();
}

void OrnamentedObj::setFrameBrushColor(QColor col)
{
	frame->setBrushColor (col);
}

QColor OrnamentedObj::getFrameBrushColor()
{
	return frame->getBrushColor ();
}

//#include <iostream>  
//using namespace std;
void OrnamentedObj::positionContents()	//FIXME-4 called multiple times for each object after moving an image with mouse
{
//cout << "OO::positionContents topPad="<<topPad<<"  botPad="<<botPad<<"  "<<treeItem->getHeadingStd()<<endl;
	double d=frame->getPadding()/2;
	double x=absPos.x();
	double y=absPos.y();

	double ox,oy;	// Offset due to padding

	ox=leftPad + d;
	oy=topPad + d;
	
	systemFlags-> move (ox +x , oy + y );

	// vertical align heading to bottom
    heading->move (ox + x + systemFlags->getBBox().width(),
				   oy + y + ornamentsBBox.height() - heading->getHeight() 
					);
	standardFlags->move (ox +x + heading->getWidth() + systemFlags->getBBox().width() , oy + y );

	ornamentsBBox.moveTopLeft ( QPointF (ox+x,oy+y));
	clickBox.moveTopLeft (QPointF (ox + x, oy + y));
}

void OrnamentedObj::move (double x, double y)
{
	MapObj::move (x,y);
	positionContents();
	updateLinkGeometry();
	requestReposition();
}

void OrnamentedObj::move (QPointF p)
{
	move (p.x(), p.y());
}	

void OrnamentedObj::moveBy (double x, double y)
{

	MapObj::moveBy (x,y);
    frame->moveBy (x,y);
    systemFlags->moveBy (x,y);
    standardFlags->moveBy (x,y);
    heading->moveBy (x,y);
	updateLinkGeometry();
	requestReposition();
}

void OrnamentedObj::moveBy (QPointF p)
{
	moveBy (p.x(), p.y());
}	

void OrnamentedObj::move2RelPos(double x, double y)
{
	setRelPos (QPointF(x,y));
	if (parObj)
	{
		QPointF p=parObj->getChildPos();
		move (p.x()+x, p.y() +y);
	}
}

void OrnamentedObj::move2RelPos(QPointF p)
{
	move2RelPos (p.x(),p.y());
}

void OrnamentedObj::activateStandardFlag(Flag *flag)
{
	standardFlags->activate(flag);
	calcBBoxSize();
	positionBBox();
	move (absPos.x(), absPos.y() );
	forceReposition();
}

void OrnamentedObj::deactivateStandardFlag(const QString &name)
{
	standardFlags->deactivate(name);
	calcBBoxSize();
	positionBBox();
	move (absPos.x(), absPos.y() );
	forceReposition();
}


QString OrnamentedObj::getSystemFlagName(const QPointF &p) //FIXME-3
{
	return systemFlags->getFlagName(p);	
}

