#include "vymview.h"

#include <iostream>
using namespace std;

#include "mainwindow.h"
#include "mapeditor.h"
#include "treeeditor.h"

extern Main *mainWindow;


VymView::VymView(VymModel *m)
{
	model=m;

	// Create TreeView
	treeEditor=new TreeEditor (model);
	//treeEditor->setModel ((QAbstractItemModel*)model);
	//treeEditor->setMinimumWidth (50);

	treeEditor->setColumnWidth (0,150);
	treeEditor->setAnimated (true);

	
	selModel=treeEditor->selectionModel();
	model->setSelectionModel (selModel);
	connect (
		selModel, SIGNAL (selectionChanged(const QItemSelection &, const QItemSelection &)), 
		this,SLOT (changeSelection(const QItemSelection &,const QItemSelection &)));

	// Create good old MapEditor
	mapEditor=model->getMapEditor();
	if (!mapEditor) mapEditor=new MapEditor (model);
	connect (
		selModel, SIGNAL (selectionChanged(const QItemSelection &, const QItemSelection &)), 
		mapEditor,SLOT (updateSelection(const QItemSelection &,const QItemSelection &)));

	connect (
		model, SIGNAL (dataChanged(const QModelIndex &, const QModelIndex &)), 
		mapEditor,SLOT (updateData(const QModelIndex &) ) );

	// VymModel may want to update selection or other data, e.g. during animation
	connect (
		model, SIGNAL (selectionChanged(const QItemSelection &, const QItemSelection &)), 
		mapEditor,SLOT (updateSelection(const QItemSelection &,const QItemSelection &)));

	connect (
		model, SIGNAL (noteHasChanged(QModelIndex) ),
		mainWindow, SLOT (updateNoteEditor (QModelIndex) ) );
		
	connect (
		model, SIGNAL (expandAll() ),
		this, SLOT (expandAll () ) );
		
	connect (
		model, SIGNAL (showSelection() ),
		this, SLOT (showSelection() ) );
		

	mapEditor->setAntiAlias (mainWindow->isAliased());
	mapEditor->setSmoothPixmap(mainWindow->hasSmoothPixmapTransform());

	addWidget (treeEditor);
	addWidget (mapEditor);

/*
	tv=new QTreeView;
	QSortFilterProxyModel *proxyModel = new QSortFilterProxyModel (this);
	proxyModel->setDynamicSortFilter (true);
	proxyModel->setSourceModel(model);
	//proxyModel->setFilterFixedString ("a",Qt::CaseInsensitive,QRegExp::FixedString);
	proxyModel->setFilterRegExp(QRegExp("e", Qt::CaseInsensitive));
	proxyModel->setFilterKeyColumn(0);
	tv->setModel (proxyModel);
	addWidget (tv);
*/
	// Set geometry
	QList <int> widths;
	widths<<200;
	widths<<600;
//	widths<<200;
	setSizes(widths);
}

VymView::~VymView()
{
	//cout << "Destructor VymView\n";
}

VymModel* VymView::getModel()
{
	return model;
}

MapEditor* VymView::getMapEditor()
{
	return mapEditor;
}

void VymView::initFocus()
{
	mapEditor->setFocus();
}

QItemSelectionModel* VymView::selectionModel() 
{
	if (treeEditor) 
		return selModel;
	else 
		std::cout <<"VymView::selectionModel: hey, no treeEditor so far???\n";
	return NULL;
}

void VymView::changeSelection (const QItemSelection &newsel, const QItemSelection &oldsel)
{
	// Notify mainwindow to update satellites like NoteEditor, if needed (model==currenModel...)
	mainWindow->changeSelection (model,newsel,oldsel);	// FIXME-5 maybe connect VymModel <-> MainWindow directly?
	// would require to also get current model in mainWindow

	//showSelection();
}

void VymView::expandAll()
{
	treeEditor->expandAll();
	//tv->expandAll();
}

void VymView::showSelection()
{
	QModelIndex ix=model->getSelectedIndex();
	treeEditor->scrollTo( ix, QAbstractItemView::EnsureVisible);
	mapEditor->scrollTo ( ix);
}

