#include <QGraphicsScene>
#include "xlinkitem.h"

#include "branchitem.h"
#include "linkablemapobj.h"
#include "vymmodel.h"
#include "xlinkobj.h"

#include <iostream>
using namespace std;

/////////////////////////////////////////////////////////////////
// XLinkItem
/////////////////////////////////////////////////////////////////

XLinkItem::XLinkItem (const QList<QVariant> &data, TreeItem *parent):MapItem (data,parent)

{
	//cout << "Const XLinkItem () "<<this<<endl;
	init();
}

XLinkItem::~XLinkItem ()
{
	cout << "Destr XLinkItem "<<this<<"  lmo="<<lmo<<endl;
	if (lmo){cout <<" calling delete\n"; delete (lmo);}
	if (partnerXLink)
	{
		// Also delete partner 
		cout << "  deleting partner="<<partnerXLink<<endl;
		partnerXLink->partnerXLink=NULL;	// avoid endless recusion
		model->deleteItem (partnerXLink);
	}
}


void XLinkItem::init () 
{
	setType (XLink);
	beginBranch=NULL;
	endBranch=NULL;
	partnerXLink=NULL;
	isBeginXLink=true;
	xLinkState=XLinkItem::undefinedXLink;

	color=QColor (180,180,180);
	width=1;
}

void XLinkItem::setBegin (BranchItem *bi)
{
	if (bi) 
	{
		xLinkState=initXLink;
		beginBranch=bi;
	}	
}

BranchItem* XLinkItem::getBegin ()
{
	return beginBranch;
}

void XLinkItem::setEnd (BranchItem *bi)
{
	if (bi) 
	{
		xLinkState=initXLink;
		endBranch=bi;
	}		
}

BranchItem* XLinkItem::getEnd()
{
	return endBranch;
}

void XLinkItem::setWidth (int w)
{
	if (isBeginXLink)
	{
		width=w;
		if (lmo) ((XLinkObj*)lmo)->updateXLink();
		return;
	}
	if (partnerXLink)
		partnerXLink->setWidth (w);
}

int XLinkItem::getWidth()
{
	if (isBeginXLink) return width;
	if (partnerXLink)
		return partnerXLink->getWidth();
	else
		return -1;
}

void XLinkItem::setColor(QColor c)
{
	if (isBeginXLink)
	{
		color=c;
		if (lmo) ((XLinkObj*)lmo)->updateXLink();
		return;
	}	
	if (partnerXLink)
		partnerXLink->setColor (c);
}

QColor XLinkItem::getColor()
{
	if (isBeginXLink) return color;
	if (partnerXLink)
		return partnerXLink->getColor();
	else
		return QColor();
}

void XLinkItem::setEnd (QPointF p)
{
	if (lmo) ((XLinkObj*)lmo)->setEnd (p);
}

bool XLinkItem::activate ()	
{
	if (beginBranch && endBranch)
	{
		if (beginBranch==endBranch) return false;

		partnerXLink=model->createXLink (endBranch);
		partnerXLink->setBegin (beginBranch);
		partnerXLink->setEnd (endBranch);
		partnerXLink->partnerXLink=this;
		partnerXLink->isBeginXLink=false;

		xLinkState=activeXLink;
		partnerXLink->xLinkState=activeXLink;
		partnerXLink->setHeading ("xLink to: "+beginBranch->getHeading());
		setHeading ("xLink to: "+endBranch->getHeading());

		model->updateActions();
		return true;
	} else
		return false;
}

bool XLinkItem::isBegin()
{
	return isBeginXLink;
}

void XLinkItem::updateXLink()
{
	if(lmo && isBeginXLink) 
		((XLinkObj*)lmo)->updateXLink();
	else 
		if (partnerXLink) partnerXLink->updateXLink();
}

void XLinkItem::updateVisibility()
{
	if (lmo) lmo->updateVisibility();
}

BranchItem* XLinkItem::getPartnerBranch()
{
	if (!beginBranch && !endBranch)
		return NULL;
	if (isBeginXLink)
		return endBranch;
	else	
		return beginBranch;
}


XLinkItem* XLinkItem::getPartnerXLink()
{
	return partnerXLink;
}


QString XLinkItem::saveToDir ()
{
	QString s="";
	if (beginBranch && endBranch && xLinkState==activeXLink)
	{
		if (beginBranch==endBranch )
			qWarning ("XLI::saveToDir  beginBranch==endBranch"); //FIXME-3	s=""
		else
		{
			QString colAttr=attribut ("color",color.name());
			QString widAttr=attribut ("width",QString().setNum(width,10));
			QString begSelAttr=attribut ("beginID",model->getSelectString(beginBranch));
			QString endSelAttr=attribut ("endID",  model->getSelectString(endBranch));
			s=beginElement ("xlink", colAttr +widAttr +begSelAttr +endSelAttr);

			s+=endElement ("xlink");
		}
	}
	return s;
}

XLinkObj* XLinkItem::createMapObj(QGraphicsScene *scene)  
{
	XLinkObj* xlo=new XLinkObj (scene,this);
	lmo=(LinkableMapObj*)xlo;
	return xlo;
}

