#include "xlinkobj.h"

#include "branchobj.h"
#include "branchitem.h"
#include "xlinkitem.h"

#include <iostream>
using namespace std;

/////////////////////////////////////////////////////////////////
// XLinkObj
/////////////////////////////////////////////////////////////////

int XLinkObj::arrowSize=10;					// make instances

XLinkObj::XLinkObj (QGraphicsScene* scene, TreeItem* ti):MapObj(scene,ti)
{
	//cout << "Const XLinkObj (s)\n";
	init();
}


XLinkObj::~XLinkObj ()
{
	//cout << "Destr XLinkObj\n";
	delete (line);
	delete (poly);
}


void XLinkObj::init () 
{
	XLinkItem *xli=(XLinkItem*)treeItem;
	visBranch=NULL;

	pen.setColor ( xli->getColor() );
	pen.setWidth ( xli->getWidth() );
	pen.setCapStyle (  Qt::RoundCap );
	line=scene->addLine(QLineF(1,1,1,1),pen);
    line->setZValue (Z_XLINK);
	poly=scene->addPolygon(QPolygonF(),pen, xli->getColor());
    poly->setZValue (Z_XLINK);
	setVisibility (true);
}

void XLinkObj::setEnd (QPointF p)
{
	endPos=p;
}


void XLinkObj::updateXLink()
{
	QPointF a,b;
	QPolygonF pa;
	if (visBranch)
	{
		// Only one of the linked branches is visible
		BranchObj *bo=(BranchObj*)(visBranch->getLMO());
		if (!bo) return;

		a=b=bo->getChildPos();
		if (bo->getOrientation()==LinkableMapObj::RightOfCenter)
		{
			b.setX (b.x()+25);
			
			pa.clear();
			pa<< QPointF(b.x(),b.y())<<
				QPointF(b.x()-arrowSize,b.y()-arrowSize)<<
				QPointF(b.x()-arrowSize,b.y()+arrowSize);
			poly->setPolygon(pa);
		} else
		{
			b.setX (b.x()-25);
			pa.clear();
			pa<< QPointF(b.x(),b.y())<<
				QPointF(b.x()+arrowSize,b.y()-arrowSize)<<
				QPointF(b.x()+arrowSize,b.y()+arrowSize);
			poly->setPolygon (pa);
		}	
	} else
	{
		// Both linked branches are visible
		BranchItem *bi=((XLinkItem*)treeItem)->getBegin();
		if ( bi)
		{
			// If a link is just drawn in the editor,
			// we have already a beginBranch
			BranchObj *bo=(BranchObj*)(bi->getLMO());
			if (bo)	
				a=bo->getChildPos();
			else 
				return;	
		}	
		else
			// This shouldn't be reached normally...
			a=beginPos;

		// FIXME-3 try to get rid of xLinkstate if (xLinkState==activeXLink && endBranch)
		bi=((XLinkItem*)treeItem)->getEnd();
		if (bi)
		{
			BranchObj *bo=(BranchObj*)(bi->getLMO());
			if (bo)	
				b=bo->getChildPos();
			else 
				return;	
		}
		else
			b=endPos;
	}

	beginPos=a;
	endPos=b;
	XLinkItem *xli=(XLinkItem*)treeItem;
	pen.setColor ( xli->getColor() );
	pen.setWidth ( xli->getWidth() );
	poly->setBrush (xli->getColor() );
	line->setPen (pen);
	line->setLine(a.x(), a.y(), b.x(), b.y());
}

void XLinkObj::positionBBox()
{
}

void XLinkObj::calcBBoxSize()
{
}

void XLinkObj::setVisibility (bool b)
{
	MapObj::setVisibility (b);
	if (b)
	{
		line->show();
		if (visBranch) 
			poly->show();
		else	
			poly->hide();
	}	
	else
	{
		line->hide();
		poly->hide();
	}	
}

void XLinkObj::setVisibility ()
{
	BranchItem* beginBI=((XLinkItem*)treeItem)->getBegin();
	BranchObj* beginBO=NULL;
	if (beginBI) beginBO=(BranchObj*)(beginBI->getLMO());

	BranchObj* endBO=NULL;
	BranchItem* endBI=((XLinkItem*)treeItem)->getEnd();
	if (endBI) endBO=(BranchObj*)(endBI->getLMO());
	if (beginBO && endBO)
	{
		if(beginBO->isVisibleObj() && endBO->isVisibleObj())
		{	// Both ends are visible
			visBranch=NULL;
			setVisibility (true);
		} else
		{
			if(!beginBO->isVisibleObj() && !endBO->isVisibleObj())
			{	//None of the ends is visible
				visBranch=NULL;
				setVisibility (false);
			} else
			{	// Just one end is visible, draw a symbol that shows
				// that there is a link to a scrolled branch
				if (beginBO->isVisibleObj())
					visBranch=beginBI;
				else
					visBranch=endBI;
				setVisibility (true);
			}
		}
	}
}


