#include "xml-vym.h"

#include <QMessageBox>
#include <QColor>
#include <QTextStream>
#include <iostream>
#include <typeinfo>

#include "branchitem.h"
#include "misc.h"
#include "settings.h"
#include "linkablemapobj.h"
#include "mainwindow.h"
#include "version.h"
#include "xlinkitem.h"


extern Main *mainWindow;
extern Settings settings;
extern QString vymVersion;

bool parseVYMHandler::startDocument()
{
    errorProt = "";
    state = StateInit;
    laststate = StateInit;
	stateStack.clear();
	stateStack.append(StateInit);
	htmldata="";
	isVymPart=false;
    return true;
}

bool parseVYMHandler::startElement  ( const QString&, const QString&,
                    const QString& eName, const QXmlAttributes& atts ) 
{
    QColor col;
	/* Testing
	cout << "startElement <"<< qPrintable(eName)
		<<">  state="<<state 
		<<"  laststate="<<stateStack.last()
		<<"   loadMode="<<loadMode
	//	<<"       line="<<QXmlDefaultHandler::lineNumber()
		<<endl;
	*/	
	stateStack.append (state);	
    if ( state == StateInit && (eName == "vymmap")  ) 
	{
        state = StateMap;
		branchesTotal=branchesCurrent=0;

		if (loadMode==NewMap )
		{
			// Create mapCenter
			model->clear();
			lastBranch=NULL;

			if (!atts.value( "author").isEmpty() )
				model->setAuthor(atts.value( "author" ) );
			if (!atts.value( "comment").isEmpty() )
				model->setComment (atts.value( "comment" ) );
			if (!atts.value( "branchCount").isEmpty() )
			{
				branchesTotal=atts.value("branchCount").toInt();
				if (branchesTotal>10)
				{
					mainWindow->setProgressMinimum (0);
					mainWindow->setProgressMaximum (branchesTotal);
					mainWindow->setProgressValue(0);
				}
			} else
			{
				mainWindow->setProgressMinimum (0);
				mainWindow->setProgressMaximum (0);
				mainWindow->setProgressValue(0);
			}

				
			if (!atts.value( "backgroundColor").isEmpty() )
			{
				col.setNamedColor(atts.value("backgroundColor"));
				model->getScene()->setBackgroundBrush(col);
			}	    
			if (!atts.value( "selectionColor").isEmpty() )
			{
				col.setNamedColor(atts.value("selectionColor"));
				model->setSelectionColor(col);
			}	    
			if (!atts.value( "linkColorHint").isEmpty() ) 
			{
				if (atts.value("linkColorHint")=="HeadingColor")
					model->setMapLinkColorHint(LinkableMapObj::HeadingColor);
				else
					model->setMapLinkColorHint(LinkableMapObj::DefaultColor);
			}
			if (!atts.value( "linkStyle").isEmpty() ) 
				model->setMapLinkStyle(atts.value("linkStyle"));
			if (!atts.value( "linkColor").isEmpty() ) 
			{
				col.setNamedColor(atts.value("linkColor"));
				model->setMapDefLinkColor(col);
			}	
			if (!atts.value( "defXLinkColor").isEmpty() ) 
			{
				col.setNamedColor(atts.value("defXLinkColor"));
				model->setMapDefXLinkColor(col);
			}	
			if (!atts.value( "defXLinkWidth").isEmpty() ) 
				model->setMapDefXLinkWidth(atts.value("defXLinkWidth").toInt ());
		}	
		// Check version
		if (!atts.value( "version").isEmpty() ) 
		{
			if (!checkVersion(atts.value("version")))
				QMessageBox::warning( 0, "Warning: Version Problem" ,
				   "<h3>Map is newer than VYM</h3>"
				   "<p>The map you are just trying to load was "
				   "saved using vym " +atts.value("version")+". "
				   "The version of this vym is " + vymVersion + 
				   ". If you run into problems after pressing "
				   "the ok-button below, updating vym should help.");
			else	   
				model->setVersion(atts.value( "version" ));

		}

	} else if ( eName == "select" && state == StateMap ) 
	{
		state=StateMapSelect;
	} else if ( eName == "setting" && state == StateMap ) 
	{
		state=StateMapSetting;
		if (loadMode==NewMap)
			readSettingAttr (atts);
	} else if ( eName == "mapcenter" && state == StateMap ) 
	{
		state=StateMapCenter;
		if (loadMode==NewMap)
		{	
			// Really use the found mapcenter as MCO in a new map
			lastBranch=model->createMapCenter(); 
		} else
		{
			// Treat the found mapcenter as a branch 
			// in an existing map
			BranchItem *bi=model->getSelectedBranch();	
			cout << "xml-vym  bi="<<bi->getHeadingStd()<<"  loadMode="<<loadMode<<endl;
			if (bi)
			{
				lastBranch=bi;
				if (loadMode==ImportAdd)
				{
					lastBranch=model->createBranch(lastBranch);
				} //else
					model->clearItem(lastBranch); 
			} else
				// add mapCenter without parent
				lastBranch=model->createMapCenter(); 
		}		
		readBranchAttr (atts);
	} else if ( 
		(eName == "standardflag" ||eName == "standardFlag") && 
		(state == StateMapCenter || state==StateBranch)) 
	{
		state=StateStandardFlag;
	} else if ( eName == "heading" && (state == StateMapCenter||state==StateBranch)) 
	{
		laststate=state;
		state=StateHeading;
		if (!atts.value( "textColor").isEmpty() ) 
		{
			col.setNamedColor(atts.value("textColor"));
			lastBranch->setHeadingColor(col );
		}	    
	} else if ( eName == "note" && 
				(state == StateMapCenter ||state==StateBranch))
	{	// only for backward compatibility (<1.4.6). Use htmlnote now.
		state=StateNote;
		if (!readNoteAttr (atts) ) return false;
	} else if ( eName == "htmlnote" && state == StateMapCenter) 
	{
		laststate=state;
		state=StateHtmlNote;
    } else if ( eName == "floatimage" && 
				(state == StateMapCenter ||state==StateBranch)) 
	{
		state=StateImage;
		lastImage=model->createImage(lastBranch);
		if (!readImageAttr(atts)) return false;
	} else if ( (eName == "branch"||eName=="floatimage") && state == StateMap) 
	{
		// This is used in vymparts, which have no mapcenter or for undo
		isVymPart=true;
		TreeItem *ti=model->getSelectedItem();
		if (!ti)
		{
			// If a vym part is _loaded_ (not imported), 
			// selection==lmo==NULL
			// Treat it like ImportAdd then...
			loadMode=ImportAdd;
			// FIXME-3 lmo=model->first()->getLMO();		
			// Do we really have no MCO when loading?????
			cout << "xml-vym aborted\n";
			return false;
		}	
		if (ti && ti->isBranchLikeType() )
		{
			lastBranch=(BranchItem*)ti;
			if (eName=="branch")
			{
				state=StateBranch;
				if (loadMode==ImportAdd)
				{
					lastBranch=model->createBranch(lastBranch);
				} else
					model->clearItem (lastBranch);
				readBranchAttr (atts);
			} else if (eName=="floatimage")
			{
				state=StateImage;
				lastImage=model->createImage (lastBranch);
				if (!readImageAttr(atts)) return false;
			} else return false;
		} else return false;
	} else if ( eName == "branch" && state == StateMapCenter) 
	{
		state=StateBranch;
		lastBranch=model->createBranch(lastBranch);
		readBranchAttr (atts);
	} else if ( eName == "htmlnote" && state == StateBranch) 
	{
		laststate=state;
		state=StateHtmlNote;
		no.clear();
		if (!atts.value( "fonthint").isEmpty() ) 
			no.setFontHint(atts.value ("fonthint") );
	} else if ( eName == "frame" && (state == StateBranch||state==StateMapCenter)) 
	{
		laststate=state;
		state=StateFrame;
		if (!readFrameAttr(atts)) return false;
    } else if ( eName == "xlink" && state == StateBranch ) 
	{
		state=StateBranchXLink;
		if (!readXLinkAttr (atts)) return false;
    } else if ( eName == "branch" && state == StateBranch ) 
	{
		lastBranch=model->createBranch(lastBranch);
		readBranchAttr (atts);
    } else if ( eName == "html" && state == StateHtmlNote ) 
	{
		state=StateHtml;
		htmldata="<"+eName;
		readHtmlAttr(atts);
		htmldata+=">";
    } else if ( state == StateHtml ) 
	{
		// accept all while in html mode,
		htmldata+="<"+eName;
		readHtmlAttr(atts);
		htmldata+=">";
    } else
        return false;   // Error
    return true;
}

bool parseVYMHandler::endElement  ( const QString&, const QString&, const QString &eName)
{
	/* Testing
	cout << "endElement </" <<qPrintable(eName)
		<<">  state=" <<state 
	//	<<"  laststate=" <<laststate
	//	<<"  stateStack="<<stateStack.last() 
	//	<<"  selString="<<model->getSelectString().toStdString()
		<<endl;
	*/
    switch ( state ) 
	{
		case StateMap:
			mainWindow->removeProgressBar();
			break;
        case StateMapCenter: 
			model->emitDataHasChanged (lastBranch);
			lastBranch=(BranchItem*)(lastBranch->parent());
		//	lastBranch->setLastSelectedBranch (0);	// Reset last selected to first child branch
            break;
        case StateBranch: 
			// Empty branches may not be scrolled 
			// (happens if bookmarks are imported)
			if (lastBranch->isScrolled() && lastBranch->branchCount()==0) 
				lastBranch->unScroll();
			model->emitDataHasChanged (lastBranch);

			lastBranch=(BranchItem*)(lastBranch->parent());
			lastBranch->setLastSelectedBranch (0);	// Reset last selected to first child branch
            break;
        case StateHtml: 
			htmldata+="</"+eName+">";
			if (eName=="html")
			{
				state=StateHtmlNote;  
				htmldata.replace ("<br></br>","<br />");
				no.setNote (htmldata);
				lastBranch->setNoteObj (no);
			}	
			break;
		default: 
			break;
    }  
	state=stateStack.takeLast();	
	return true;
}

bool parseVYMHandler::characters   ( const QString& ch)
{
	//cout << "characters \""<<ch.toStdString()<<"\"  state="<<state <<"  laststate="<<laststate<<endl;

	QString ch_org=quotemeta (ch);
    QString ch_simplified=ch.simplifyWhiteSpace();
    if ( ch_simplified.isEmpty() ) return true;

    switch ( state ) 
    {
        case StateInit: break;
        case StateMap: break; 
		case StateMapSelect:
			model->select(ch_simplified);
			break;
		case StateMapSetting:break;
        case StateMapCenter: break;
        case StateNote:
			lastBranch->setNote(ch_simplified);
			break;
        case StateBranch: break;
        case StateStandardFlag: 
            lastBranch->activateStandardFlag(ch_simplified); 
            break;
        case StateImage: break;
        case StateHtmlNote: break;
        case StateHtml:
			htmldata+=ch_org;
			break;
        case StateHeading: 
            lastBranch->setHeading(ch_simplified);
            break;
        default: 
			return false;
    }
    return true;
}

QString parseVYMHandler::errorString() 
{
    return "the document is not in the VYM file format";
}

bool parseVYMHandler::readBranchAttr (const QXmlAttributes& a)	
{
	mainWindow->setProgressValue (branchesCurrent++);

	lastMI=lastBranch;

	if (!readOOAttr(a)) return false;

	if (!a.value( "scrolled").isEmpty() )
		lastBranch->toggleScroll();	
		// (interesting for import of KDE bookmarks)

/*	if (!a.value( "frameType").isEmpty() )  FIXME-3
		lastOO->setFrameType (a.value("frameType")); //Compatibility 1.8.1

*/	
	if (!a.value( "incImgV").isEmpty() ) 
	{	
		if (a.value("incImgV")=="true")
			lastBranch->setIncludeImagesVer(true);
		else	
			lastBranch->setIncludeImagesVer(false);
	}	
	if (!a.value( "incImgH").isEmpty() ) 
	{	
		if (a.value("incImgH")=="true")
			lastBranch->setIncludeImagesHor(true);
		else	
			lastBranch->setIncludeImagesHor(false);
	}	
	return true;	
}

bool parseVYMHandler::readFrameAttr (const QXmlAttributes& a)	// FIXME-4 does not work if there is no lmo for treeitem
{
	if (lastMI)
	{
		OrnamentedObj* oo=(OrnamentedObj*)(lastMI->getLMO()); 
		if (oo)
		{
			bool ok;
			int x;
			{
				if (!a.value( "frameType").isEmpty() ) 
					oo->setFrameType (a.value("frameType"));
				if (!a.value( "penColor").isEmpty() ) 
					oo->setFramePenColor (a.value("penColor"));
				if (!a.value( "brushColor").isEmpty() ) 
					oo->setFrameBrushColor (a.value("brushColor"));
				if (!a.value( "padding").isEmpty() ) 
				{
					x=a.value("padding").toInt(&ok);
					if (ok) oo->setFramePadding(x);
				}	
				if (!a.value( "borderWidth").isEmpty() ) 
				{
					x=a.value("borderWidth").toInt(&ok);
					if (ok) oo->setFrameBorderWidth(x);
				}	
			}		
			return true;
		}
	}
	return false;
}

bool parseVYMHandler::readOOAttr (const QXmlAttributes& a)
{
	if (lastMI)
	{
		bool okx,oky;
		float x,y;
		if (!a.value( "relPosX").isEmpty() ) 
		{
			if (!a.value( "relPosY").isEmpty() ) 
			{
				x=a.value("relPosX").toFloat (&okx);
				y=a.value("relPosY").toFloat (&oky);
				if (okx && oky  )
					lastMI->setRelPos (QPointF(x,y));
				else
					return false;   // Couldn't read relPos
			}           
		}           
		if (!a.value( "absPosX").isEmpty() ) 
		{
			if (!a.value( "absPosY").isEmpty() ) 
			{
				x=a.value("absPosX").toFloat (&okx);
				y=a.value("absPosY").toFloat (&oky);
				if (okx && oky  )
					lastMI->setAbsPos (QPointF(x,y));
				else
					return false;   // Couldn't read absPos
			}           
		}           
		if (!a.value( "id").isEmpty() ) 
			lastMI->setID (a.value ("id"));		
			
		if (!a.value( "url").isEmpty() ) 
			lastMI->setURL (a.value ("url"));
		if (!a.value( "vymLink").isEmpty() ) 
			lastMI->setVymLink (a.value ("vymLink"));
		if (!a.value( "hideInExport").isEmpty() ) 
			if (a.value("hideInExport")=="true")
				lastMI->setHideInExport(true);

		if (!a.value( "hideLink").isEmpty()) 
		{
			if (a.value ("hideLink") =="true")
				lastMI->setHideLinkUnselected(true);
			else	
				lastMI->setHideLinkUnselected(false);
		}	
	}
	return true;	
}

bool parseVYMHandler::readNoteAttr (const QXmlAttributes& a)
{	// only for backward compatibility (<1.4.6). Use htmlnote now.
	no.clear();
	QString fn;
	if (!a.value( "href").isEmpty() ) 
	{
		// Load note
		fn=parseHREF(a.value ("href") );
		QFile file (fn);
		QString s;						// Reading a note

		if ( !file.open( QIODevice::ReadOnly) )
		{
			qWarning ("parseVYMHandler::readNoteAttr:  Couldn't load "+fn);
			return false;
		}	
		QTextStream stream( &file );
		QString lines;
		while ( !stream.atEnd() ) {
			lines += stream.readLine()+"\n"; 
		}
		file.close();

		lines ="<html><head><meta name=\"qrichtext\" content=\"1\" /></head><body>"+lines + "</p></body></html>";
		no.setNote (lines);
	}		
	if (!a.value( "fonthint").isEmpty() ) 
		no.setFontHint(a.value ("fonthint") );
	lastBranch->setNoteObj(no);
	return true;
}

bool parseVYMHandler::readImageAttr (const QXmlAttributes& a)
{
	lastMI=lastImage;
	
	//if (!readOOAttr(a)) return false;   FIXME-3

	if (!a.value( "href").isEmpty() )
	{
		// Load Image
		if (!lastImage->load (parseHREF(a.value ("href") ) ))
		{
			QMessageBox::warning( 0, "Warning: " ,
				"Couldn't load image\n"+parseHREF(a.value ("href") ));
			lastImage=NULL;
			return true;
		}
		
	}	
	if (!a.value( "zPlane").isEmpty() ) 
		lastImage->setZValue (a.value("zPlane").toInt ());
    float x,y;
    bool okx,oky;
	if (!a.value( "relPosX").isEmpty() ) 
	{
		if (!a.value( "relPosY").isEmpty() ) 
		{
			// read relPos
			x=a.value("relPosX").toFloat (&okx);
			y=a.value("relPosY").toFloat (&oky);
			if (okx && oky) 
				lastImage->setRelPos (QPointF (x,y) );
			else
				// Couldn't read relPos
				return false;  
		}           
	}	
	
	//FIXME-3 if (!readOOAttr(a)) return false;

	if (!a.value ("originalName").isEmpty() )
	{
		lastImage->setOriginalFilename (a.value("originalName"));
	}
	return true;
}

bool parseVYMHandler::readXLinkAttr (const QXmlAttributes& a)	
{
	// object ID is used starting in version 1.8.76
	// (before there was beginBranch and endBranch)
	if (!a.value( "beginID").isEmpty() ) 
	{ 
		if (!a.value( "endID").isEmpty() ) 
		{
			TreeItem *beginBI=model->findID (a.value( "beginID"));
			TreeItem   *endBI=model->findID (a.value( "endID"));
			if (beginBI && endBI && beginBI->isBranchLikeType() && endBI->isBranchLikeType() )
			{
				XLinkItem *xli=model->createXLink (lastBranch,true);
				xli->setBegin ( (BranchItem*)beginBI );
				xli->setEnd ( (BranchItem*)endBI);
				xli->activate();

				if (!a.value( "color").isEmpty() ) 
				{
					QColor col;
					col.setNamedColor(a.value("color"));
					xli->setColor (col);
				}

				if (!a.value( "width").isEmpty() ) 
				{
					bool okx;
					xli->setWidth(a.value ("width").toInt (&okx, 10));
				}
				xli->updateXLink();
			}
		}           
	}	
	return true;	// xLinks can only be established at the "end branch", return true
}

bool parseVYMHandler::readHtmlAttr (const QXmlAttributes& a)
{
	for (int i=1; i<=a.count(); i++)
		htmldata+=" "+a.localName(i-1)+"=\""+a.value(i-1)+"\"";
	return true;
}

bool parseVYMHandler::readSettingAttr (const QXmlAttributes& a)
{
	if (!a.value( "key").isEmpty() ) 
	{
		if (!a.value( "value").isEmpty() ) 
			settings.setLocalEntry (model->getDestPath(), a.value ("key"), a.value ("value"));
		else
			return false;
		
	} else
		return false;
	
	return true;
}
